
-- why not use _ENV anyway lol
self = self or {}

LUA_HINT_NONE = 0
LUA_HINT_ARRAY = 1
LUA_HINT_OBJECT = 2

-- this replicates original engine code, but it shouldn't work in first place
local function __newindex(self, key, value)
	local nils = getmetatable(self).__nils

	-- If we are setting an entry to nil, need to add a bogus integer entry
	-- to the __nils table, otherwise need to set the entry *in* the __nils
	-- table to nil and remove it.

	-- TODO: __newindex is called only when assigning non-existing keys to values,
	-- TODO: as per Lua manual.
	-- TODO: Chucklefish weren't aware of this?

	if key == nil then
		nils[key] = true
	else
		nils[key] = nil
	end

	rawset(self, key, value)
end

function jobject()
	return setmetatable({}, {
		__newindex = __newindex,
		__nils = {},
		__typehint = LUA_HINT_OBJECT
	})
end

function jarray()
	return setmetatable({}, {
		__newindex = __newindex,
		__nils = {},
		__typehint = LUA_HINT_ARRAY
	})
end

function jremove(self, key)
	if type(self) ~= 'table' then error('bad argument #1 to jremove: table expected, got ' .. type(self), 2) end
	local meta = getmetatable(self)

	if meta and meta.__nils then
		meta.__nils[key] = true
	end

	self[key] = nil
end

function jsize(self)
	if type(self) ~= 'table' then error('bad argument #1 to jsize: table expected, got ' .. type(self), 2) end

	local elemCount = 0
	local highestIndex = 0
	local hintList = false

	local meta = getmetatable(self)

	if meta then
		if meta.__typehint == LUA_HINT_ARRAY then
			hintList = true
		end

		local nils = meta.__nils

		if nils ~= nil then
			for k, v in pairs(nils) do
				if type(k) == 'number' then
					highestIndex = math.max(k, highestIndex)
				else
					hintList = false
				end

				elemCount = elemCount + 1
			end
		end
	end

	for k, v in pairs(self) do
		if type(k) == 'number' then
			highestIndex = math.max(k, highestIndex)
		else
			hintList = false
		end

		elemCount = elemCount + 1
	end

	if hintList then
		return highestIndex
	else
		return elemCount
	end
end

function jresize(self, target)
	if type(self) ~= 'table' then error('bad argument #1 to jresize: table expected, got ' .. type(self), 2) end

	local meta = getmetatable(self)

	if meta and meta.__nils then
		local indices = {}

		for k, v in pairs(meta.__nils) do
			if type(k) == 'number' and k % 1.0 == 0.0 and k > target then
				table.insert(indices, k)
			end
		end

		for i, v in ipars(indices) do
			meta.__nils[v] = nil
		end
	end

	local indices = {}

	for k, v in pairs(self) do
		if type(k) == 'number' and k % 1.0 == 0.0 and k > target then
			table.insert(indices, k)
		end
	end

	for i, v in ipars(indices) do
		self[v] = nil
	end
end

local __print = __print

function print(...)
	local values = {...}

	for i, v in ipairs(values) do
		values[i] = tostring(v)
	end

	__print(table.concat(values, '\t'))
end
